from __future__ import annotations

import json
from json import JSONDecodeError
from threading import Lock
from typing import Any, Optional

from .constants import CONFIG_FILENAME
from .exceptions import AuthenticationError


class ConfigMeta(type):
    _instances = {}  # type: ignore
    _lock: Lock = Lock()  # type: ignore

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
            return cls._instances[cls]


class Config(metaclass=ConfigMeta):
    """Config helper class."""

    config_properties: list = [
        'api_key'
    ]

    def __init__(self):
        """Constructs Config object."""
        self._api_key: Optional[str] = None

    # def set_api_key(self, api_key=Optional[None]):
    #     """set api_key"""
    #     if api_key:
    #         self._api_key = api_key
    #     else:
    #         raise APIKeyNotSupplied()

    @property
    def api_key(self) -> str:
        """API key property of the Config object.

        Returns:
            str: API key.
        """
        if self._api_key:
            return self._api_key
        else:
            try:
                with open(CONFIG_FILENAME) as config_read:
                    self._api_key = json.load(config_read)['api_key']
                    return self._api_key
            except FileNotFoundError:
                raise AuthenticationError()
            except JSONDecodeError:
                raise

    @api_key.setter
    def api_key(self, value):
        self._api_key = value

    def save(self):
        """Saves configuration to the JSON config file."""
        with open(CONFIG_FILENAME, 'w') as config_write:
            # TODO FUTURE: select only specific keys
            config_load = {k: getattr(self, k)
                           for k in Config.config_properties}
            json.dump(config_load, config_write)

    def __str__(self) -> str:
        """Returns structured configuration information."""
        return str(self.__dict__)



import json
import pytest

from dynacrop import Polygon
# from .test_utils import supply_polygon


def get_mocked_data(path):
    with open(path) as md:
        return json.load(md)


@pytest.fixture(autouse=True)
def refresh_mocker(mocker):
    mocker.patch(
        'dynacrop.api_handles.APIObject.refresh',
        return_value=None
    )


@pytest.fixture(autouse=True)
def get_mocker(mocker, path):
    mocker.patch(
        'dynacrop.api_handles.RequestsHelper.get',
        return_value=get_mocked_data(path)
    )


@pytest.fixture(autouse=True)
def patch_mocker(mocker):
    mocker.patch(
        'dynacrop.api_handles.RequestsHelper.patch',
        return_value=None
    )


@pytest.fixture()
def example_apiobject(apiobject_type):
    return apiobject_type.get(1)


# @pytest.fixture
# def example_polygon(mocker):
#     mocker.patch(
#         'dynacrop.api_handles.RequestsHelper.get',
#         return_value=get_mocked_data('mocked_polygon_data.json')
#     )
#     mocker.patch(
#         'dynacrop.api_handles.RequestsHelper.patch',
#         return_value=None
#     )
#     return Polygon.get(1)


# @pytest.fixture
# def polygons_list(mocker):
#     mocker.patch(
#         'dynacrop.api_handles.RequestsHelper.get',
#         return_value=get_mocked_data('mocked_list_polygons.json')
#     )
#     mocker.patch(
#         'dynacrop.api_handles.APIObject.refresh',
#         return_value=None
#     )
#     return Polygon.list()

# @pytest.fixture
# def deleted_polygon(mocker):



import os
from enum import Enum

# TODO constants to Config
BASEURL = 'https://api-dynacrop.worldfromspace.cz/api/v2'
MAX_WAIT_TIME = 30
CONFIG_PATH: str = os.path.join(
        os.path.expanduser('~'),
        '.dynacrop'
    )
CONFIG_FILENAME = os.path.join(CONFIG_PATH, 'config.json')

# Create config folder
if not os.path.isdir(CONFIG_PATH):
    os.mkdir(CONFIG_PATH)


class CRS(Enum):
    EPSG4326 = 'EPSG:4326'


class Layer(Enum):
    NDVI = 'NDVI'
    EVI = 'EVI'
    FAPAR = 'FAPAR'
    LAI = 'LAI'
    NDMI = 'NDMI'
    CCC = 'CCC'
    CWC = 'CWC'
    NDWI = 'NDWI'
    MSAVI2 = 'MSAVI2'
    NDRE = 'NDRE'
    NDREX = 'NDREX'
    SMI = 'SMI'
    IRECI = 'IRECI'
    NDDI = 'NDDI'
    NMDI = 'NMDI'
    MNDWI = 'MNDWI'
    WIW = 'WIW'


class RenderingType(Enum):
    OBSERVATION = 'observation'
    TIME_SERIES = 'time_series'
    FIELD_ZONATION = 'field_zonation'
    FIELD_ZONATION_BY_MEDIAN = 'field_zonation_by_median'


class Result(Enum):
    RAW = "raw"
    COLOR = "color"
    PNG = "png"
    TILES_COLOR = "tiles_color"
    TILES_DEMO = "tiles_demo"
    SHP = "shp"
    GEOJSON = "geojson"
    SHP_SIMPLIFIED = "shp_simplified"
    GEOJSON_SIMPLIFIED = "geojson_simplified"
    JSON = "json"
    STATISTICS = "statistics"
    FREQUENCIES = "frequencies"



import math
import os
import shutil
from datetime import date, timedelta
from functools import partial
from random import choice, randrange, uniform

import pyproj
import requests
from dynacrop.processing_request_base import ProcessingRequest
from dynacrop import (FieldZonation, Config, Layer,
                      FieldZonationByMedian, Observation, TimeSeries)
from shapely.geometry import Point
from shapely.ops import transform


def authenticate():
    config = Config()
    config.api_key = 'api_key'


def supply_polygon() -> str:
    """Constructs and supplies a small random polygon within central European
    coordinates.

    Returns:
        str: Polygon in Well-Known Text.
    """
    size = 1000
    proj = partial(
        pyproj.transform,
        pyproj.Proj('epsg:4326'),
        pyproj.Proj('epsg:3857')
    )
    proj_back = partial(
        pyproj.transform,
        pyproj.Proj('epsg:3857'),
        pyproj.Proj('epsg:4326')
    )
    centre_point = Point(uniform(15, 16), uniform(48, 50))
    centre_point_3857 = transform(proj, centre_point)
    sample_polygon = centre_point_3857.buffer(math.sqrt(size) / 2).envelope
    sample_polygon_4326 = transform(proj_back, sample_polygon)
    return sample_polygon_4326.wkt


def supply_layer() -> Layer:
    """Supplies a random layer from DynaCrop Layers.

    Returns:
        Layer: Random DynaCrop Layer
    """
    return choice([layer for layer in list(Layer) if layer != Layer.SMI])


def supply_dates() -> tuple:
    """Supplies random dates from 1 to 3 months apart within Sentinel
    satellite operation years.

    Returns:
        tuple: consists of date from and date to
    """
    # Get random year within Sentinel operation
    random_year = randrange(2018, 2022)
    # Prepare range of the whole year for random date
    start_date = date(random_year, 1, 1)
    end_date = date(random_year, 12, 31)
    # Select random number (seed) of days from the prepared range
    days_between_dates = (end_date - start_date).days
    random_number_of_days = randrange(days_between_dates)

    # Create date to (add this random number to start date)
    date_to = start_date + timedelta(days=random_number_of_days)

    # Create date from (subtract 1 up to 3 months from date to)
    date_from = date.fromordinal(date_to.toordinal() - randrange(30, 120))

    return str(date_from), str(date_to)


def fetch_results(proc_req: ProcessingRequest):
    """Fires and checks all possible
    result outputs including saving them to temporary folder.
    Decides appropriately based on processing
    request type.

    Args:
        proc_req (ProcessingRequest): One of the DynaCrop processing requests.
    """
    # Creates temporary folder for checking storing
    # results
    temp_result_dir = './tmp_healthcheck'
    if not os.path.isdir(temp_result_dir):
        os.mkdir(temp_result_dir)

    if isinstance(proc_req, TimeSeries):
        proc_req.get_json()
        proc_req.save_json('tmp_healthcheck/healthcheck_result.json')
    else:
        proc_req.as_array()
        proc_req.as_geodataframe()
        proc_req.save_tiff('tmp_healthcheck/healthcheck_result.tiff')
        proc_req.save_colored_tiff(
            'tmp_healthcheck/healthcheck_result_colored.tiff')
        proc_req.save_png('tmp_healthcheck/healthcheck_result.png')
        proc_req.save_shapefile('tmp_healthcheck/healthcheck_result.shp')
        proc_req.save_geojson('tmp_healthcheck/healthcheck_result.geojson')
        requests.head(proc_req.get_demo_tiles_url()).raise_for_status()

        if isinstance(proc_req, FieldZonation) \
           or isinstance(proc_req, FieldZonationByMedian):
            proc_req.save_simplified_shapefile(
                'tmp_healthcheck/healthcheck_result.shp')
            proc_req.save_simplified_geojson(
                'tmp_healthcheck/healthcheck_result.geojson')

        proc_req.get_frequencies()

        if isinstance(proc_req, Observation):
            proc_req._get_stats()

    # Removes temporary folder
    if os.path.isdir(temp_result_dir):
        shutil.rmtree(temp_result_dir)



# For typehinting
# Available only for Python 3.7+
from __future__ import annotations

import json
import webbrowser
from datetime import datetime
from json.decoder import JSONDecodeError

import fiona  # type: ignore
import geopandas as gpd  # type: ignore
import numpy as np  # type: ignore
import requests  # type: ignore
from rasterio.io import MemoryFile  # type: ignore

from .api_handles import APIObject
from .attrs import ProcessingRequestAttrs
from .constants import Result
from .exceptions import (InvalidRenderingTypeError,
                         PolygonNotReadyForProcessingError, RequestHasNoData)


class ProcessingRequest(APIObject, ProcessingRequestAttrs):
    """Processing request endpoint object"""
    _editable_attrs: set = set()
    _endpoint: str = 'processing_request'
    _allowed_result_types: set = set()

    def refresh(self):
        """Updates the response data of the Processing request.

        Raises:
            InvalidRenderingTypeError: If an already-created processing
                request is a different rendering type than requested.
        """
        super(ProcessingRequest, self).refresh()
        if ('rendering_type' in self._data.keys() and
                self._data['rendering_type'] != self.rendering_type):
            raise InvalidRenderingTypeError(
                self._data['id'],
                self._data['rendering_type'])

    @classmethod
    def create(cls, **kwargs) -> APIObject:
        """Creates Processing request endpoint object.

        Returns:
            APIObject: APIObject instatiated into ProcessingRequest child.
        """
        cls.validate(kwargs)
        kwargs['date_from'], kwargs['date_to'] = \
            ProcessingRequest.correct_dates(
                [kwargs['date_from'],
                 kwargs['date_to']]
            )

        return super(ProcessingRequest, cls).create(**kwargs)

    @classmethod
    def validate(cls, kwargs: dict):
        """Validates Processing request result status.

        Args:
            kwargs (dict)

        Raises:
            PolygonNotReadyForProcessingError: If the Processing request is
                created or requested while the target Polygon is not yet
                processed.
        """
        if 'polygon' in kwargs and not kwargs['polygon'].is_ready():
            raise PolygonNotReadyForProcessingError(kwargs['polygon'].status)

    @staticmethod
    def correct_dates(dates: list) -> list:
        """Checks if dates need to be corrected and does so.

        Args:
            dates (list): Dates within a processing request.

        Returns:
            tuple: List with converted dates.
        """
        return (
            [ProcessingRequest.convert_date_to_str(date)
                if isinstance(date, datetime)
                else date
                for date in dates]
        )

    @staticmethod
    def convert_date_to_str(date: datetime) -> str:
        """Converts datetime.date to string if necessary

        Args:
            date (datetime): Date in datetime.datetime format.

        Returns:
            str: Date in string format.
        """
        try:
            return datetime.strftime(date, "%Y-%m-%d")
        except ValueError:
            raise

    def has_data(self) -> bool:
        """Checks whether the Processing request returned any data.

        Returns:
            bool: Data filling statetment.
        """
        return (
            (self._data['status'] != 'no_data') and
            (self._data['status'] != 'error') and
            ('result' in self._data)
        )


class RasterProcessingRequest(ProcessingRequest):
    """Baseclass for raster and vector Processing request results.
Mainly serves for handlig results."""

    def _get_result_url(self, result_type: Result) -> str:
        """Acquires result URL from the Processing request response
                based on desired result type.

        Args:
            result_type (Result): One of the available result types.
                See https://bit.ly/2Yjg6wk for further information.

        Raises:
            RequestHasNoData: If result has returned no data. This might
                be caused due to unavailability of satellite images within
                the date range.

        Returns:
            str: URL to download/view the result.
        """
        if self.has_data():
            return self._data['result'][result_type.value]
        else:
            raise RequestHasNoData(self.id)

    def _get_content(self, result_type: Result) -> bytes:
        """Acquires the content of the Processing request response.

        Args:
            result_type (Result): One of the available result types
                for the given Processing request. See https://bit.ly/2Yjg6wk
                for further information.

        Returns:
            bytes: Response content as bytes text.
        """
        return requests.get(
            self._get_result_url(result_type),
            allow_redirects=True).content

    def _get_raster(self, result_type: Result) -> np.ndarray:
        """Where applicable, returns raster result as NumPy array.

        Args:
            result_type (Result): One of the available *raster* result types
                for the given Processing request. See https://bit.ly/2Yjg6wk
                for further information.

        Returns:
            numpy.ndarray: Multidimensional NumPy array.
        """
        content = self._get_content(result_type)
        with MemoryFile(content) as memfile:
            with memfile.open() as dataset:
                data_array = dataset.read().astype('float32')
        return data_array

    def _save_raster(self, result_type: Result, path: str):
        """Saves raster result of the Processing request.

        Args:
            result_type (Result): One of the available *raster* result types
                for the given Processing request. See https://bit.ly/2Yjg6wk
                for further information.
            path (str): Path to save the file.
        """
        content = self._get_content(result_type)
        with open(path, 'wb') as dst:
            dst.write(content)

    def _get_vector(self, result_type) -> gpd.GeoDataFrame:
        """Where applicable, returns vector result as Geopandas GeoDataFrame.

        Returns:
            gpd.GeoDataFrame: Geopandas GeoDataFrame.
        """
        content = self._get_content(result_type)
        return gpd.GeoDataFrame.from_features(fiona.BytesCollection(content))

    def as_array(self, clouds_as_nan: bool = True,
                 outwards_as_nan: bool = True) -> np.ndarray:
        """Wrapper method for private _get_raster.

        Args:
            clouds_as_nan (bool): Whether convert clouds to NumPy.nan instead
                of default -999
            outwards_as_nan (bool): Whether convert values outside of the
                field shape to NumPy.nan instead of deafult -998

        Returns:
            numpy.ndarray: Multidimensional NumPy array.
        """
        raw = self._get_raster(Result.RAW)

        if clouds_as_nan:
            raw[raw == -999] = np.nan
        if outwards_as_nan:
            raw[raw == -998] = np.nan

        return raw

    # RAW
    def get_tiff_url(self) -> str:
        """Returns RAW (TIFF) URL.

        Returns:
            str: RAW (TIFF) URL.
        """
        return self._get_result_url(Result.RAW)

    def save_tiff(self, path: str):
        """Saves result in RAW TIFF format.

        Args:
            path (str): Path to save the file.
        """
        self._save_raster(Result.RAW, path)

    # COLOR
    def get_colored_tiff_url(self) -> str:
        """Returns COLORED TIFF URL.

        Returns:
            str: COLORED TIFF URL.
        """
        return self._get_result_url(Result.COLOR)

    def save_colored_tiff(self, path: str):
        """Saves result in COLORED TIFF format.

        Args:
            path (str): Path to save the file.
        """
        self._save_raster(Result.COLOR, path)

    # PNG
    def get_png_url(self) -> str:
        """Returns PNG URL.

        Returns:
            str: PNG URL.
        """
        return self._get_result_url(Result.PNG)

    def save_png(self, path: str):
        """Saves result in PNG format.

        Args:
            path (str): Path to save the file.
        """
        self._save_raster(Result.PNG, path)

    # TILES COLOR
    def get_tiles_url(self) -> str:
        """Get tiles URL.

        Returns:
            str: Tiles URL.
        """
        return self._get_result_url(Result.TILES_COLOR)

    # TILES DEMO
    def get_demo_tiles_url(self) -> str:
        """Get demo tiles URL for preview.

        Returns:
            str: Demo tiles URL.
        """
        return self._get_result_url(Result.TILES_DEMO)

    def preview(self):
        """Opens default browser in a new tab using
                demo tiles URL. This is an easy result preview.
        """
        url = self._get_result_url(Result.TILES_DEMO)
        webbrowser.open(url, new=2)

    ############
    # VECTORS
    ############
    def as_geodataframe(self) -> gpd.GeoDataFrame:
        """Wrapper method for private _get_vector.

        Returns:
            gpd.GeoDataFrame: Geopandas GeoDataFrame.
        """
        return self._get_vector(Result.SHP)

    # SHAPEFILE
    def get_shapefile_url(self) -> str:
        """Returns shapefile URL.

        Returns:
            str: Shapefile URL.
        """
        return self._get_result_url(Result.SHP)

    def save_shapefile(self, path: str):
        """Saves result in as shapefile.

        Args:
            path (str): Path to save the file.
        """
        gdf = self._get_vector(Result.SHP)
        gdf.to_file(path)

    # GEOJSON
    def get_geojson_url(self) -> str:
        """Returns GeoJSON URL.

        Returns:
            str: GeoJSON URL.
        """
        return self._get_result_url(Result.GEOJSON)

    def save_geojson(self, path: str):
        """Saves result as GeoJSON.

        Args:
            path (str): Path to save the file.
        """
        gdf = self._get_vector(Result.GEOJSON)
        gdf.to_file(path, driver='GeoJSON')

    # #####
    # FREQS
    # #####
    def get_frequencies(self) -> dict:
        """Acquires percentual shares of value intervals
        of the field area.

        Returns:
            dict: Keys are tuples of intervals, values are percentual
                shares.

        """
        freqs_bytes = self._get_content(Result.FREQUENCIES)

        # Stupid workaround because we default return bytes text
        # from _get_content...
        # Kind of doubles with get_statistics. But since it is strictly
        # in a different scope, I'd leave it.
        try:
            freqs = json.loads(freqs_bytes.decode('utf-8'))
        except JSONDecodeError:
            raise

        return {eval(k): v for k, v in freqs.items()}


class JSONProcessingRequest(ProcessingRequest):
    """Baseclass for JSON Processing request results.
Mainly serves for handlig results."""
    def get_json(self) -> dict:
        """Acquires JSON result as a dictionary from the Processing request response.

        Raises:
            RequestHasNoData: If result has returned no data. This might
                be caused due to unavailability of satellite images within
                the date range.

        Returns:
            dict: JSON data as Python dictionary.
        """
        if self.has_data():
            return self._data['result']['time_series']
        else:
            raise RequestHasNoData(self.id)

    def save_json(self, path: str):
        """Saves JSON result.

        Args:
            path (str): Path to save the file.
        """
        content = self.get_json()
        with open(path, 'w') as dst:
            json.dump(content, dst)


class ObservationBase(RasterProcessingRequest):
    """Baseclass for specific methods of Observation"""

    def _get_stats(self) -> dict:
        """Acquires descriptive statistics that are always present
        for the Observation endpoint for the current field.
        This method is should not be used itself. It is encapsulated in
        methods regarding particular stats. You can choose from:
        *mean*, *median*, *standard deviation*, *minimum*, *maximum*
        and dictionary of *percentiles*.

        Returns:
            dict: Dictionary of yet unselected statistics.
        """
        stats_bytes = self._get_content(Result.STATISTICS)

        # Stupid workaround because we return bytes text
        # from _get_content by default
        try:
            return json.loads(stats_bytes.decode('utf-8'))
        except JSONDecodeError:
            raise

    def get_mean(self) -> float:
        """Returns mean of the field from the Observation endpoint.

        Returns:
            float: Mean of the field values.
        """
        return self._get_stats()['mean']

    def get_median(self) -> float:
        """Returns median of the field from the Observation endpoint.

        Returns:
            float: Median of the field values.
        """
        return self._get_stats()['median']

    def get_sd(self) -> float:
        """Returns standard deviation of the field from the Observation endpoint.

        Returns:
            float: Standard deviation of the field values.
        """
        return self._get_stats()['sd']

    def get_min(self) -> float:
        """Returns minimum of the field from the Observation endpoint.

        Returns:
            float: Minimum of the field values.
        """
        return self._get_stats()['min']

    def get_max(self) -> float:
        """Returns maximum of the field from the Observation endpoint.

        Returns:
            float: Maximum of the field values.
        """
        return self._get_stats()['max']

    def get_percentiles(self) -> dict:
        """Returns percentiles of the field from the Observation endpoint.

        Returns:
            float: Percentiles of the field values.
        """
        return self._get_stats()['percentiles']


class FieldZonationBase(RasterProcessingRequest):
    """Baseclass for specific methods of FieldZonation"""
    # SIMPLIFIED GEOJSON
    def get_simplified_geojson_url(self) -> str:
        """Returns simplified GeoJSON URL.

        Returns:
            str: Simplified GeoJSON URL.
        """
        return self._get_result_url(Result.GEOJSON_SIMPLIFIED)

    def save_simplified_geojson(self, path: str):
        """Saves simplified GeoJSON result.

        Args:
            path (str): Path to save the file.
        """
        gdf = self._get_vector(Result.GEOJSON_SIMPLIFIED)
        gdf.to_file(path, driver='GeoJSON')

    # SIMPLIFIED SHP
    def get_simplified_shapefile_url(self) -> str:
        """Returns simplified shapefile URL.

        Returns:
            str: Simplified shapefile URL.
        """
        return self._get_result_url(Result.SHP_SIMPLIFIED)

    def save_simplified_shapefile(self, path: str):
        """Saves simplified shapefile result.

        Args:
            path (str): Path to save the file.
        """
        gdf = self._get_vector(Result.SHP_SIMPLIFIED)
        gdf.to_file(path)



class AuthenticationError(Exception):
    def __init__(self):
        super().__init__("Wrong or missing API key.")


class APIKeyNotSupplied(Exception):
    def __init__(self):
        super().__init__("API key not supplied.")


class HandlingResponseError(Exception):
    def __init__(self):
        super().__init__("""There was a unexpected problem processing the
        request.""")


class APIObjectNotFound(Exception):
    def __init__(self, url):
        super().__init__(f"API Object on {url} not found.")


class APIRequestNotValid(Exception):
    def __init__(self, error):
        super().__init__(f"API request is not valid: {error}.")


class UnexpectedServerError(Exception):
    def __init__(self):
        super().__init__("API encoutered unexpected "
                         "problems when processing the request.")


class ServiceNotImplementedError(Exception):
    def __init__(self):
        super().__init__("API does not recognize the requested service.")


class BadGatewayError(Exception):
    def __init__(self):
        super().__init__("API recieved an invalid response \
            from the upstream server.")


class ServiceUnavailableError(Exception):
    def __init__(self):
        super().__init__("DynaCrop API is temporarily unavailable.\
            Please try again later.")


class InvalidLayerSuppliedError(Exception):
    def __init__(self):
        super().__init__("The supplied layer is in invalid format. \
            Use constants.Layer instead.")


class PolygonNotReadyForProcessingError(Exception):
    def __init__(self, status):
        super().__init__(f"The polygon is not yet ready for processing. \
            Polygon status: {status}.")


class InvalidRenderingTypeError(Exception):
    def __init__(self, id, rendering_type):
        super().__init__(f"The processing request with id {id} is \
            of different type: {rendering_type}.")


class NotInAllowedResultTypesError(Exception):
    def __init__(self, rendering_type, allowed_result_types):
        super().__init__(f"Invalid result type for {rendering_type}. \
            Allowed result types are: \
            {[art.value for art in allowed_result_types]}.")


class InvalidResultSuppliedError(Exception):
    def __init__(self):
        super().__init__("The supplied result type is in invalid format. \
        Use constants.Result instead.")


class RequestHasNoData(Exception):
    def __init__(self, id):
        super().__init__(f"The request with id={id} has no data.")


class MissingID(Exception):
    def __init__(self, id):
        super().__init__("The object has no id.")


class IncorrectDateFormatError(Exception):
    def __init__(self):
        super().__init__("""Date from or date to are in incorrect format. Must
        be string of only date in ISO 8601""")




from typing import Optional

from ..constants import Layer
from .async_request import AsyncRequestAttrs


class ProcessingRequestAttrs(AsyncRequestAttrs):
    date_from: str
    date_to: str
    value_no_data: int
    value_clouds: int
    result: dict
    layer: Layer
    polygon_id: int
    number_of_zones: Optional[int]
    thresholds: list



from typing import Iterator
import pytest

from dynacrop import Polygon


@pytest.mark.parametrize(
    'path, apiobject_type', [
        ('mocked_polygon_data.json', Polygon)
    ],
)
def test_correct_instance(example_apiobject, apiobject_type):
    assert isinstance(example_apiobject, apiobject_type)


@pytest.mark.parametrize(
    'path, apiobject_type',
    [
        ('mocked_polygon_data.json', Polygon)
    ]
)
def test_apiobject_has_id(example_apiobject):
    assert example_apiobject.id


@pytest.mark.parametrize('path, apiobject_type', [
    ('mocked_polygon_data.json', Polygon)]
)
@pytest.mark.parametrize('noneditable_attr', [
    'id',
    'geometry',
    'area',
    'last_valid_observation',
    'valid_observations',
    'cloud_cover_percent',
    'last_updated']
)
def test_polygon_noneditable_attrs(
    example_apiobject,
    noneditable_attr
):
    with pytest.raises(AttributeError):
        setattr(example_apiobject, noneditable_attr, 'nonsense')


# @pytest.mark.parametrize(
#     'polygon_attr, attr_value',
#     zip(
#         Polygon._editable_attrs,
#         [True, 0.8, '[CHANGED LABEL] DynaCrop SDK Polygon for tests']
#     )
# )
# def test_polygon_editable_attrs(example_polygon, polygon_attr, attr_value):
#     setattr(example_polygon, polygon_attr, attr_value)


# def test_list_polygons(polygons_list):
#     assert isinstance(polygons_list, Iterator)


# def test_list_polygons_correct_results(polygons_list):
#     assert all([isinstance(p, Polygon) for p in polygons_list])



from .async_request import AsyncRequestAttrs


class PolygonAttrs(AsyncRequestAttrs):
    geometry: str
    area: int
    last_valid_observation: str
    smi_enabled: bool
    max_mean_cloud_cover: float
    label: str
    valid_observations: list
    cloud_cover_percent: list
    last_updated: str



class UserAttrs:
    username: str
    id: int
    email: str
    api_key: str
    current_plan: str
    current_plan_start_date: str
    current_plan_end_date: str
    polygon_count: int
    max_number_of_polygons: int
    max_calls_per_minute: int
    remaining_calls_per_minute: int
    remaining_processing_capacity: int



from .config import Config  # noqa
from .constants import CRS, Layer, RenderingType, Result  # noqa
from .sdk import (FieldZonation, FieldZonationByMedian, Observation,  # noqa
                  Polygon, TimeSeries, User)
from .processing_request_base import ProcessingRequest  # noqa



from __future__ import annotations

from time import sleep
from typing import Any, Optional, Set

import requests

from .config import Config
from .constants import BASEURL
from .exceptions import (AuthenticationError, APIObjectNotFound,
                         APIRequestNotValid,
                         BadGatewayError, HandlingResponseError, MissingID,
                         ServiceNotImplementedError, ServiceUnavailableError,
                         UnexpectedServerError)


class RequestsHelper:
    """Interface for handling requests."""

    @classmethod
    def get(cls, id: str, endpoint: str) -> dict:
        """Creates a GET request for the supplied id and endpoint.

        Args:
            id (str): Used to indentify the endpoint instance.
            endpoint (str): One of the DynaCrop API endpoints.

        Returns:
            dict: Reponse to the request.
        """

        r = requests.get(
            cls.build_url(endpoint, id),
            params={'api_key': Config().api_key})
        return cls.handle_response(r)

    @classmethod
    def patch(cls, id: str, data: dict, endpoint: str) -> dict:
        """Creates a PATCH request for the supplied id, endpoint
    and patch data.

        Args:
            id (str): Used to indentify the endpoint instance.
            data (dict): Used to patch the existing backend data.
            endpoint (str): One of the DynaCrop API endpoints.


        Returns:
            dict: Response to the request with patched data.
        """
        r = requests.patch(
            cls.build_url(endpoint, id),
            params={'api_key': Config().api_key}, data=data)
        return cls.handle_response(r)

    @classmethod
    def post(cls, data: dict, endpoint: str) -> dict:
        """Creates a POST request for the supplied id, endpoint and post data.

        Args:
            id (str): Used to indentify the endpoint instance.
            data (dict): Used to create new instance of the endpoint
            on the backend.
            endpoint (str): One of the DynaCrop API endpoints.


        Returns:
            dict: Response to the request with a new instance
            of endpoint.
        """
        r = requests.post(
            cls.build_url(endpoint),
            params={'api_key': Config().api_key},
            data=data)
        return cls.handle_response(r)

    @classmethod
    def delete(cls, id: str, endpoint: str):
        """Creates a DELETE request for the supplied endpoint and id.

        Args:
            id (str): Used to indentify the endpoint instance.
            endpoint (str): One of the DynaCrop API endpoints.

        Returns:
            [type]: [description]
        """
        r = requests.delete(
            cls.build_url(endpoint, id),
            params={'api_key': Config().api_key})
        return cls.handle_response(r)

    @classmethod
    def list(
        cls,
        endpoint: Optional[str] = None,
        url: Optional[str] = None
    ) -> dict:
        """Fetches bulk version of an endpoint to return pagination info and
        multiple endpoint results.

        Args:
            endpoint (str): One of the DynaCrop API endpoints. Either endpoint
                or url must be stated.
            url (int): URL to list endpoints. Either endpoint
                or url must be stated.

        Returns:
            dict: Response to the request.
        """
        if endpoint:
            req = cls.build_url(endpoint)
            params = {'api_key': Config().api_key}
        elif url:
            req = url
            params = {}
        else:
            # TODO custom exception?
            raise Exception

        r = requests.get(
            req,
            params=params)
        return cls.handle_response(r)

    @classmethod
    def handle_response(cls, response: requests.Response):
        """Handles response from the request methods.

        Args:
            response (dict): Reponse supplied via request methods
            of the RequestHelper class

        Raises:
            APIObjectNotFound: 404 Not Found. Might be a non-existent
                response content or a API-side error.
            APIRequestNotValid: 400 Bad Request. Might be a wrongly
                defined request.
            UnexpectedServerError: 500 Internal Server Error. Might be an API
                error that can only be resolved on server side.
            ServiceNotImplementedError: 501 Not Implemented. The request was
                not recognized by the API server. Might be implemented
                in the future.
            BadGatewayError: 502 Bad Gateway. There could have been an error
                beyond the DynaCrop API.
            ServiceUnavailableError: 503 Service Unavailable. The API
                is temporarily unavailable.
            HandlingResponseError: Some other unrecognized error.

        Returns: None
        """
        if response.status_code == 200 or response.status_code == 201:
            return response.json()
        elif response.status_code == 204:
            pass
        elif response.status_code == 401:
            raise AuthenticationError()
        elif response.status_code == 404:
            raise APIObjectNotFound(response.url)
        elif response.status_code == 400:
            raise APIRequestNotValid(response.json())
        elif response.status_code == 500:
            raise UnexpectedServerError
        elif response.status_code == 501:
            raise ServiceNotImplementedError
        elif response.status_code == 502:
            raise BadGatewayError
        elif response.status_code == 503:
            raise ServiceUnavailableError
        else:
            raise HandlingResponseError

    @classmethod
    def build_url(cls, *args: Optional[str]) -> str:
        """Builds internet url based on parameters

        Returns:
            str: Built URL.
        """
        return '/'.join([BASEURL] + [str(x) for x in args])


class APIObject:
    """Metaclass for various endpoints of the API."""
    _editable_attrs: Set[Any] = set()
    _endpoint: str = ""
    _object_attrs: Set = {'_data'}

    def __init__(self, id: Optional[int] = None):
        """Constructs an API object.

        Args:
            id (str): ID of an endpoint request.
        """
        self._data: dict = {'id': str(id) if id else ''}
        self.refresh()

    def __dir__(self) -> list:
        return list(self._data.keys())

    def __getattr__(self, item: str) -> Any:
        """An override for getattr.

        Args:
            item (str): Attribute to be found from the object

        Raises:
            AttributeError: If attribute not found.

        Returns:
            Any: APIObject's attribute.
        """

        if self._data['id']:
            try:
                return self._data[item]
            except KeyError:
                raise AttributeError(item)
        else:
            raise Exception(f'{self} has no data.')

    def __setattr__(self, key: str, value: str):
        """An override for setattr.

        Args:
            key (str): The name of the attribute.
            value (str): The value of the attribute.

        Raises:
            AttributeError: If accessed a non-editable attribute .
        """

        if key in self._object_attrs:
            super(APIObject, self).__setattr__(key, value)
        else:
            if key in self._editable_attrs:
                self._data[key] = value
                self._data = RequestsHelper.patch(
                    self.id,
                    {key: self._data[key]
                        for key
                        in self._editable_attrs},
                    self._endpoint)
            else:
                raise AttributeError(key)

    @classmethod
    def get(cls, id: Optional[int] = None) -> APIObject:
        """Acquires APIObject instatiated into one of the DynaCrop API endpoints.

        Args:
            id (int): ID of an endpoint request.

        Returns:
            APIObject: APIObject made to an API endpoint.
        """
        return cls(id)

    @classmethod
    def create(cls, **kwargs) -> APIObject:
        """Creates APIObject later instatiated into one of the DynaCrop API endpoints.

        Args:
            kwargs (Optional[str])

        Returns:
            APIObject: APIObject made to an API endpoint.
        """
        return cls.get(RequestsHelper.post(
            {key: val for key, val in kwargs.items() if val},
            cls._endpoint)['id'])

    def delete(self):
        """Deletes APIObject from the API.
        """
        RequestsHelper.delete(self.id, self._endpoint)
        self._data = {'id': ''}

    @classmethod
    def list(cls) -> APIObjectIterator:
        """Fetches APIObjects in bulk as an iterator.

        Returns:
            APIObjectIterator: Iterator of APIObjects.
        """
        return iter(APIObjectIterator(cls))

    def is_ready(self) -> bool:
        """Checks whether the request is ready.

        Returns:
            bool: Is ready statement.
        """
        self.refresh()

        return self._data['status'] != 'created' \
            and self._data['status'] != 'rendering'

    def is_failed(self) -> bool:
        """Checks whether the request failed.

        Returns:
            bool: Failure statement.
        """
        return self._data['status'] == 'error'

    def block_till_completed(self, polling_interval: int = 1):
        """Suspends the code execution until the request response is
    returned as finished/with error/with no data.

        Args:
            polling_interval (int, optional): Time to wait between
            checking iterations. Defaults to 1.
        """
        while not self.is_ready():
            sleep(polling_interval)

    def refresh(self):
        """Updates response data."""
        self._data = RequestsHelper.get(self._data['id'], self._endpoint)
        if 'id' not in self._data:
            raise MissingID

    def __eq__(self, other: object) -> bool:
        """Compares two instantiated endpoints (APIObjects).

        Args:
            other (APIObject): Another APIObject to compare.

        Returns:
            bool: Comparation statement.
        """
        if isinstance(other, APIObject):
            return other.id == self.id
        return NotImplemented

    def __repr__(self) -> str:
        """Returns structured information about APIObject.

        Returns:
            str: Structured information about APIObject.
        """
        if self._data and 'id' in self._data:
            return f"<dynacrop.{self.__class__.__name__} {self._data['id']}>"
        else:
            return f"<dynacrop.{self.__class__.__name__} no_id>"


# TODO Make it work for specific endpoints
class APIObjectIterator:
    """Iterator for listing existing APIObjects.
    """
    def __init__(self, cls: APIObject):
        """Constructs iterator.

        Args:
            cls (APIObject): APIObject for recognizing further iterator
                operations.
        """
        self.cls = cls
        self.page = 1
        self.object_index = 0
        self.apiobject_list = RequestsHelper.list(
            endpoint=self.cls._endpoint)

    def __iter__(self) -> APIObjectIterator:
        """Returns iterator.

        Returns:
            APIObjectIterator: The iterator.
        """
        return self

    def __next__(self) -> APIObject:
        """Acquires next object on the page within pagination. Automatically
        calls next page if necessary.

        Raises:
            StopIteration: When there no more results to iterate over.

        Returns:
            APIObject: APIObject made to an API endpoint.
        """
        try:
            endpoint_data = self.apiobject_list['results'][self.object_index]
            self.object_index += 1
            return self.cls(endpoint_data['id'])
        except IndexError:
            self.object_index = 0
            if 'next' in self.apiobject_list['pagination']:
                self.apiobject_list = RequestsHelper.list(
                    url=self.apiobject_list['pagination']['next'])
                endpoint_data = \
                    self.apiobject_list['results'][self.object_index]
                self.object_index += 1
                return self.cls(endpoint_data['id'])
            else:
                raise StopIteration



from setuptools import find_packages, setup  # type: ignore
VERSION = '0.0.2-1'
LONG_DESCRIPTION = 'Python Software Development Kit library for DynaCrop API'

setup(
    name="dynacrop",
    version=VERSION,
    author='World from Space',
    description="Python Software Development Kit library for DynaCrop API",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        'geopandas>=0.7.0',
        'rasterio>=1.0a12',
        'numpy>=1.20.0',
        'requests>=2.20.0',
        'Fiona>=1.8.13',

    ]
)



from __future__ import annotations

from typing import Any, Optional, Set

from .api_handles import APIObject
from .attrs import PolygonAttrs, UserAttrs
from .constants import Layer
from .processing_request_base import (FieldZonationBase, JSONProcessingRequest,
                                      ObservationBase)


class User(APIObject, UserAttrs):
    """User endpoint object"""
    _editable_attrs: Set[Any] = set()
    _endpoint: str = 'user'

    def __init__(self):
        """Creates User endpoint object"""
        super(User, self).__init__()

    def __repr__(self) -> dict:
        """Prints information about the user.

        Returns:
            str: The information.
        """
        return str(self.__dict__['_data'])


class Polygon(APIObject, PolygonAttrs):
    """Polygon endpoint object"""
    _editable_attrs: Set[Any] = {
        'max_mean_cloud_cover',
        'label',
        'smi_enabled'}
    _endpoint: str = 'polygons'

    @classmethod
    def create(cls,  # type: ignore
               geometry: str,
               label: Optional[str] = None,
               max_mean_cloud_cover: Optional[int] = None,
               smi_enabled: bool = False, **kwargs) -> Polygon:  # type: ignore
        """Creates Polygon endpoint object.

        Args:
            geometry (str): Valid polygon shape in
                Well Known Text (WKT) representation
                (see: https://en.wikipedia.org/wiki/Well-\
                    known_text_representation_of_geometry).
            label (Optional[str], optional): Description to the Polygon
                (user field). Defaults to None.
            max_mean_cloud_cover (Optional[int], optional): Maximum mean cloud
                coverage in decimal percentage (i.e. 0.3). Defaults to None.
            smi_enabled (bool, optional): To enable Soil Moisture Index
                pre-computation. Defaults to False.

        Returns:
            Polygon: APIObject instantiated into Polygon child.
        """
        return super(Polygon, cls).create(
                geometry=geometry,
                label=label,
                max_mean_cloud_cover=max_mean_cloud_cover,
                smi_enabled=smi_enabled)  # type: ignore


class TimeSeries(JSONProcessingRequest):
    """Time series endpoint object. Time series is the service of
DynaCrop API."""
    rendering_type = 'time_series'

    @classmethod
    def create(cls,  # type: ignore
               polygon: Polygon,
               layer: Layer,
               date_from: str,
               date_to: str) -> TimeSeries:
        """Creates Time series endpoint object.

        Args:
            polygon (Polygon): Polygon endpoint object.
            layer (str): Should be Layer enumration. One of the DynaCrop API
                layers. See https://dynacrop.worldfromspace.cz/docs/#/products
                for further information.
            date_from (str): Date to record time series from.
            date_to (str): Date to record time series to.

        Returns:
            APIObject: APIObject instatiated into TimeSeries child.
        """
        return super(TimeSeries, cls).create(polygon_id=polygon.id,
                                             rendering_type=cls.rendering_type,
                                             layer=layer.value,
                                             date_from=date_from,
                                             date_to=date_to)  # type: ignore


class Observation(ObservationBase):
    """Observation endpoint object. Observation is the service of
DynaCrop API."""
    rendering_type = 'observation'

    @classmethod
    def create(cls,  # type: ignore
               polygon: Polygon,
               layer: Layer,
               date_from: str,
               date_to: str) -> Observation:  # type: ignore
        """Creates Observation endpoint object.

        Args:
            polygon (Polygon): Polygon endpoint object.
            layer (str): Should be Layer enumration. One of the DynaCrop API
                layers. See https://dynacrop.worldfromspace.cz/docs/#/products
                for further information.
            date_from (str): Date to watch for observation from.
            date_to (str): Date to watch for observation to.

        Returns:
            APIObject: APIObject instatiated into Observation child.
        """
        return super(Observation, cls). \
            create(
                polygon_id=polygon.id,
                rendering_type=cls.rendering_type,
                layer=layer.value,
                date_from=date_from,
                date_to=date_to)  # type: ignore


class FieldZonation(FieldZonationBase):
    """Field zonation endpoint object. Field zonation is the service of
    DynaCrop API."""
    rendering_type = 'field_zonation'

    @classmethod
    def create(cls,  # type: ignore
               polygon: Polygon,
               layer: Layer,
               date_from: str,
               date_to: str,
               number_of_zones: int = 3) -> FieldZonation:  # type: ignore
        """Creates Field zonation endpoint object.

        Args:
            polygon (Polygon): Polygon endpoint object.
            layer (str): Should be Layer enumration. One of the DynaCrop API
                layers. See https://dynacrop.worldfromspace.cz/docs/#/products
                for further information.
            date_from (str): Date to compute field zonation from.
            date_to (str): Date to compute field zonation to.
            number_of_zones (int): Number of zones to separate the field to.
                The number of zones must be one of 3, 5, 10, 20, 255.

        Returns:
            APIObject: APIObject instatiated into FieldZonation child.
        """
        return super(FieldZonation, cls). \
            create(
                polygon_id=polygon.id,
                rendering_type=cls.rendering_type,
                layer=layer.value,
                date_from=date_from,
                date_to=date_to,
                number_of_zones=str(number_of_zones))  # type: ignore


class FieldZonationByMedian(FieldZonationBase):
    """Field zonation by median endpoint object. Field zonation by median
is the service of DynaCrop API."""
    rendering_type = 'field_zonation_by_median'

    @classmethod
    def create(cls,  # type: ignore
               polygon: Polygon,
               layer: Layer,
               date_from: str,
               date_to: str,
               thresholds: Optional[list] = None) \
            -> FieldZonationByMedian:  # type: ignore
        """Creates Field zonation by median endpoint object.

        Args:
            polygon (Polygon): Polygon endpoint object.
            layer (str): Should be Layer enumration. One of the DynaCrop API
                layers. See https://dynacrop.worldfromspace.cz/docs/#/products
                for further information.
            date_from (str): Date to compute field zonation by median from.
            date_to (str): Date to compute field zonation by median to.
            thresholds (list): Thresholds to zone the field in between.
                The number of thresholds must be one of 2, 4, 9, 19, 254.

        Returns:
            APIObject: APIObject instatiated into FieldZonationByMedian child.
        """
        return super(FieldZonationByMedian, cls).\
            create(
                polygon_id=polygon.id,
                rendering_type=cls.rendering_type,
                layer=layer.value,
                date_from=date_from,
                date_to=date_to,
                thresholds=thresholds)  # type: ignore
